"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
var __importStar = (this && this.__importStar) || function (mod) {
    if (mod && mod.__esModule) return mod;
    var result = {};
    if (mod != null) for (var k in mod) if (Object.hasOwnProperty.call(mod, k)) result[k] = mod[k];
    result["default"] = mod;
    return result;
};
Object.defineProperty(exports, "__esModule", { value: true });
var fs_1 = __importDefault(require("fs"));
var ts = __importStar(require("typescript"));
var parser = __importStar(require("@babel/parser"));
var traverse_1 = __importDefault(require("@babel/traverse"));
var generator_1 = __importDefault(require("@babel/generator"));
var prettier_1 = __importDefault(require("prettier"));
function findNearestNode(position, ast) {
    var result;
    traverse_1.default(ast, {
        BinaryExpression: function (path) {
            var node = path.node;
            if (!result) {
                result = path;
                return;
            }
            // We are searching for the last expression that ends before the error
            /**
             * 1
             * 2
             * 3 position
             * 4 node
             */
            if ((node.start || Infinity) > position) {
                return;
            }
            /**
             * 1 node
             * 2 result
             * 3 position
             * 4
             */
            if ((node.end || Infinity) < (result.node.end || Infinity)) {
                return;
            }
            result = path;
        }
    });
    // @ts-ignore
    return result;
}
var ignoreComment = " @ts-ignore typescript-all-in";
function addIgnoreComment(path) {
    if (!path || !path.node) {
        return;
    }
    path.node.leadingComments = path.node.leadingComments || [];
    // only add if there is no comment yet
    if (!path.node.leadingComments.find(function (comment) { return comment.value === ignoreComment; })) {
        // @ts-ignore
        path.node.leadingComments.push({
            type: "CommentLine",
            value: ignoreComment
        });
    }
}
function fixErrorsForFile(diagnosticList, logger) {
    if (!diagnosticList.length) {
        return;
    }
    var file = diagnosticList[0].file;
    var fileName = file.fileName;
    // get positions
    var errorPositions = diagnosticList
        .filter(function (diagnostic) { return typeof diagnostic.start === "number"; })
        .map(function (diagnostic) { return diagnostic.start; });
    // parse the file
    var ast = parser.parse(fs_1.default.readFileSync(fileName, "utf-8"), {
        sourceType: "unambiguous",
        plugins: [
            "typescript",
            "jsx",
            "dynamicImport",
            "decorators-legacy",
            "classProperties"
        ]
    });
    logger.log("Parsed " + fileName);
    // find all ast nodes
    var errorAstPaths = errorPositions
        .map(function (position) { return findNearestNode(position, ast); })
        .filter(function (result) { return result !== null; });
    logger.log("Adding " + errorAstPaths.length + " comments to " + fileName);
    // update ast nodes
    errorAstPaths.forEach(addIgnoreComment);
    // save the file
    var code = generator_1.default(ast).code;
    var prettyCode = code;
    try {
        prettyCode = prettier_1.default.format(code, { parser: "typescript" });
    }
    catch (e) {
        logger.info("Could not prettify code for " + fileName);
    }
    logger.log("Writing " + fileName);
    fs_1.default.writeFileSync(fileName, prettyCode, "utf-8");
}
function detectAndFixTypescriptErrors(root, ignorePaths, logger) {
    var configPath = ts.findConfigFile(root, ts.sys.fileExists, "tsconfig.json");
    if (!configPath) {
        throw new Error("Could not find a valid 'tsconfig.json'.");
    }
    // tslint:disable-next-line no-any
    var host = ts.sys;
    var parsedCmd = ts.getParsedCommandLineOfConfigFile(configPath, { noEmit: true }, host);
    if (!parsedCmd) {
        throw new Error("Could not parse config");
    }
    var options = parsedCmd.options, fileNames = parsedCmd.fileNames;
    var program = ts.createProgram({
        rootNames: fileNames,
        options: options
    });
    var diagnostics = program.emit().diagnostics;
    var allDiagnostics = ts.getPreEmitDiagnostics(program).concat(diagnostics);
    var fileNameMap = {};
    allDiagnostics.forEach(function (diagnostic) {
        if (diagnostic.file) {
            var fileName = diagnostic.file.fileName;
            fileNameMap[fileName] = fileNameMap[fileName] || [];
            fileNameMap[fileName].push(diagnostic);
        }
        else {
            var msg = ts.flattenDiagnosticMessageText(diagnostic.messageText, "\n");
            logger.log("Ignoring error without file: \"" + msg + "\"");
        }
    });
    Object.entries(fileNameMap).forEach(function (_a) {
        var fileName = _a[0], diagnosticsList = _a[1];
        // shall I ignore the file?
        if (ignorePaths.some(function (ignorePath) { return fileName.includes(ignorePath); })) {
            return;
        }
        // TODO: multi-thread this approach
        fixErrorsForFile(diagnosticsList, logger);
    });
}
exports.default = detectAndFixTypescriptErrors;
//# sourceMappingURL=fixTypescriptErrors.js.map